package com.vernaio.example;

import java.io.File;
import java.util.ArrayList;
import java.util.List;

import com.vernaio.aivis.engine.v2.base.sdk.EngineLogger;
import com.vernaio.aivis.engine.v2.base.sdk.dto.DtoFilesFilter;
import com.vernaio.aivis.engine.v2.base.sdk.dto.DtoInferenceTimeseriesDataFilter;
import com.vernaio.aivis.engine.v2.base.sdk.dto.DtoInterval;
import com.vernaio.aivis.engine.v2.base.sdk.dto.DtoTimeseriesFilesReaderConfig;
import com.vernaio.aivis.engine.v2.base.sdk.dto.IDtoFloatConstraintValuesWithNextNormal;
import com.vernaio.aivis.engine.v2.base.sdk.dto.IDtoInferenceDataSpecification;
import com.vernaio.aivis.engine.v2.constraint.navigator.sdk.ConstraintNavigatorData;
import com.vernaio.aivis.engine.v2.constraint.navigator.sdk.ConstraintNavigatorHub;
import com.vernaio.aivis.engine.v2.constraint.navigator.sdk.ConstraintNavigatorInference;
import com.vernaio.aivis.engine.v2.constraint.navigator.sdk.ConstraintNavigatorModelContext;
import com.vernaio.aivis.engine.v2.constraint.navigator.sdk.ConstraintNavigatorSetup;
import com.vernaio.aivis.engine.v2.constraint.navigator.sdk.dto.DtoFloatNextNormalConfig;
import com.vernaio.aivis.engine.v2.constraint.navigator.sdk.dto.DtoHubConfig;
import com.vernaio.aivis.engine.v2.constraint.navigator.sdk.dto.DtoInferenceConfig;
import com.vernaio.aivis.engine.v2.constraint.navigator.sdk.dto.DtoL2DistanceCost;
import com.vernaio.aivis.engine.v2.constraint.navigator.sdk.dto.DtoModelConstraint;
import com.vernaio.aivis.engine.v2.constraint.navigator.sdk.dto.DtoSignalConstraint;
import com.vernaio.aivis.engine.v2.constraint.navigator.sdk.dto.DtoStaticNumericalCondition;
import com.vernaio.aivis.engine.v2.constraint.navigator.sdk.dto.IDtoHubConfig;
import com.vernaio.aivis.engine.v2.constraint.navigator.sdk.dto.IDtoModelConstraint;
import com.vernaio.aivis.engine.v2.constraint.navigator.sdk.dto.IDtoSignalConstraint;

public class Main {
	
	private static final int MIN_LOGGING_LEVEL = 3;
	
	private static final int MAX_THREADS = 0; // max threads 0 means the number of threads will be chosen automatically
	
	private static final String INFERENCE_DATA_FILE = "eval_cn.csv";
	
	public static void main(final String[] args) throws Exception {
		
		System.out.println("Setup ...");
		// parsing args as folder config
		final FolderConfig folderConfig = FolderConfig.fromArgs(args);
		// create and register logger
		ConstraintNavigatorSetup.registerLogger(new EngineLogger() {
			
			@Override
			public void log(int level, String thread, String module, String message) {
				if (level <= MIN_LOGGING_LEVEL) {
					System.out.println(String.format("\t... %s", message));
				}
			}
		});
		// init thread count
		ConstraintNavigatorSetup.initThreadCount(MAX_THREADS);
		System.out.println("\t... done!");
		
		/*
		 * MODEL CONTEXT
		 */
		
		System.out.println("Creating model context ...");
		
		// read model
		final com.vernaio.aivis.engine.v2.signal.prediction.sdk.dto.IDtoModel spModel =
			Utils.readJson(new File(folderConfig.getInputFolder(), "sp_model_0.json"), com.vernaio.aivis.engine.v2.signal.prediction.sdk.dto.IDtoModel.class);
		final ConstraintNavigatorModelContext modelContext = ConstraintNavigatorModelContext.create().addSignalPrediction("sp_model_0", spModel);
		
		System.out.println("\t... done!");
		
		/*
		 * HUB MODEL
		 */
		
		System.out.println("Creating model hub ...");
		
		// build constraint navigator hub
		final IDtoHubConfig hubConfig = new DtoHubConfig();
		final ConstraintNavigatorHub hub = ConstraintNavigatorHub.create(modelContext, hubConfig);
		
		// save hub report and hub model
		Utils.writeJson(new File(folderConfig.getOutputFolder(), "hub-model.json"), hub.getModel());
		Utils.writeJson(new File(folderConfig.getOutputFolder(), "hub-report.json"), hub.getReport());
		
		// saving memory, we don't need model context anymore
		modelContext.destroy();
		
		System.out.println("\t... done!");
		
		/*
		 * INFERENCE
		 */
		
		final long inferenceStartTime = 1749L;
		final long inferenceEndTime = 1797L;
		
		System.out.println("Loading inference data ...");
		
		final String[] files = { INFERENCE_DATA_FILE };
		final DtoFilesFilter filesFilter = new DtoFilesFilter().withIncludeFiles(files);
		
		final IDtoInferenceDataSpecification dataSpec = hub.getModel().getInferenceDataSpecification();
		final DtoInferenceTimeseriesDataFilter dataFilter =
			new DtoInferenceTimeseriesDataFilter(
				new IDtoInferenceDataSpecification[] { dataSpec },
				new DtoInterval[] { new DtoInterval().withStartTime(inferenceStartTime).withEndTime(inferenceEndTime) }
			);
		
		final DtoTimeseriesFilesReaderConfig filesReaderConfig =
			new DtoTimeseriesFilesReaderConfig(folderConfig.getInputFolder()).withFilesFilter(filesFilter).withDataFilter(dataFilter);
		
		final ConstraintNavigatorData inferenceData = ConstraintNavigatorData.create();
		inferenceData.readFiles(filesReaderConfig);
		
		System.out.println("\t... done!");
		
		System.out.println("Creating inference ...");
		
		// define model constraints
		final DtoStaticNumericalCondition modelCondition = new DtoStaticNumericalCondition().withLowerThreshold(0.9);
		final IDtoModelConstraint[] modelConstraints = { new DtoModelConstraint("sp_model_0", modelCondition) };
		
		// define signal constraints
		final IDtoSignalConstraint[] signalConstraints = new IDtoSignalConstraint[64];
		for (int index = 0; index < 64; index += 1) {
			final DtoStaticNumericalCondition signalCondition = new DtoStaticNumericalCondition().withUpperThreshold(16.0).withLowerThreshold(0.0);
			signalConstraints[index] = new DtoSignalConstraint(Long.toString(index), signalCondition);
		}
		
		// define inference config
		final DtoInferenceConfig inferenceConfig =
			new DtoInferenceConfig(true, new DtoL2DistanceCost()).withModelConstraints(modelConstraints).withSignalConstraints(signalConstraints);
		
		// note that we use here try-with-resource notation to destroy inference handle, inference.destroy() works equivalently
		try (final ConstraintNavigatorInference inference = ConstraintNavigatorInference.createByHub(hub, inferenceConfig)) {
			
			hub.destroy();
			System.out.println("\t... done!");
			
			System.out.println("Calculating next normal values ...");
			// here, we choose exactly the timestamp that are in the inference data file
			final List<Long> inferenceTimestamps = new ArrayList<Long>();
			for (long time = inferenceStartTime; time < inferenceEndTime; time += 1L) {
				inferenceTimestamps.add(time);
			}
			
			// define next normal config (needed if we want signal filter)
			final DtoFloatNextNormalConfig nextNormalConfig = new DtoFloatNextNormalConfig();
			
			// perform constraint optimization
			final List<? extends IDtoFloatConstraintValuesWithNextNormal> inferencesWithNextNormal =
				inference.inferFloatWithNextNormal(inferenceData, inferenceTimestamps, nextNormalConfig);
			
			System.out.println("\t... done!");
			
			System.out.println("Saving inference file ...");
			
			Utils.writeInferencesWithNextNormal(new File(folderConfig.getOutputFolder(), "float-constraint-values-with-next-normal.json"), inferencesWithNextNormal);
			inferenceData.destroy();
			inference.destroy();
			
			System.out.println("\t... done!");
			
		}
	}
}
