# this python script performs the same calculation as example_ad.ipynb

import json
import argparse
import os

from aivis_engine_v2_cn_sdk_python import (
    EngineLogger,
    ConstraintNavigatorSetup,
    ConstraintNavigatorData,
    ConstraintNavigatorModelContext,
    ConstraintNavigatorHub,
    ConstraintNavigatorInference,
)


# ---------------------------------------------------------------------------------
# Data location

SP_MODEL = "sp_model_0.json"
INFERENCE_DATA_FILE = "eval_cn.csv"


# Input and output folder specified via args
def from_args():
    parser = argparse.ArgumentParser(
        description="Example project for constraint navigator"
    )
    # data is expected in input folder
    parser.add_argument(
        "--input", "-i", type=str, nargs=1, required=True, help="Path to input folder"
    )
    # output files will be written to output folder
    parser.add_argument(
        "--output", "-o", type=str, nargs=1, required=True, help="Path to output folder"
    )
    args = parser.parse_args()
    # create output folder if not existing
    if not os.path.exists(args.output[0]):
        os.mkdir(args.output[0])
    return {
        "input_folder": args.input[0],
        "output_folder": args.output[0],
    }


# ---------------------------------------------------------------------------------
# Logger and Threads


class Logger(EngineLogger):
    def log(self, level, thread, module, message):
        min_logging_level = 3
        if level <= min_logging_level:
            print("\t... %s" % message)


def setup_logger_and_threads():
    print("Initializing ConstraintNavigatorSetup: logging and threads ...")
    ConstraintNavigatorSetup.register_logger(Logger())
    max_threads = 0  # number of threads will be chosen automatically with 0
    ConstraintNavigatorSetup.init_thread_count(max_threads)
    print("\t... done!")
    return


# ---------------------------------------------------------------------------------
# Reading and Writing


def read_json(filename):
    print("\t... reading file '%s' ..." % filename)
    with open(filename, "r") as file:
        return json.loads(file.read())


def write_json(filename, obj):
    print("\t... writing file '%s' ..." % filename)
    with open(filename, "w") as file:
        file.write(json.dumps(obj, indent=2))
    return


# convert [DtoFloatConstraintValuesWithNextNormal] to json
def float_constraint_values_with_next_normal_to_json(list_object, inference_config):
    return {
        "_type": "FloatConstraintValuesWithNextNormal",
        "config": json.loads(inference_config),
        "dataPoints": [
            {
                "timestamp": int(point_object.timestamp),
                "cost": {
                    "costType": str(point_object.cost.cost_type.decode("utf-8")),
                    "value": float(point_object.cost.value),
                    "nextNormalValue": float(point_object.cost.next_normal_value),
                },
                "constraints": [
                    {
                        "model": str(constraint.model.decode("utf-8")),
                        "value": float(constraint.value),
                        "nextNormalValue": float(constraint.next_normal_value),
                    }
                    for constraint in point_object.constraints
                ],
                "features": [
                    {
                        "feature": int(feature.feature),
                        "value": float(feature.value),
                        "nextNormalValue": float(feature.next_normal_value),
                        "rating": float(feature.rating),
                    }
                    for feature in point_object.features
                ],
            }
            for point_object in list_object
        ],
    }


# ---------------------------------------------------------------------------------
# MAIN


def main():
    folder_config = from_args()  # parsing args as folder_config
    setup_logger_and_threads()

    # ---- MODEL CONTEXT ----

    print("Creating model context ...")
    # initialize constriant navigator model context
    model_context = ConstraintNavigatorModelContext.create()
    # add signal prediction model to model context
    model_context.add_signal_prediction(
        "sp_model_0",
        json.dumps(read_json(os.path.join(folder_config["input_folder"], SP_MODEL))),
    )
    print("\t... done!")

    # ---- HUB MODEL ----

    print("Creating model hub ...")
    # build constraint navigator hub
    ch_config = json.dumps({})
    hub = ConstraintNavigatorHub.create(model_context, ch_config)
    # get hub model
    hub_model = json.loads(hub.get_model())
    # get hub report
    hub_report = hub.get_report()

    # save hub report and hub model
    write_json(
        os.path.join(folder_config["output_folder"], "hub-model.json"), hub_model
    )

    write_json(
        os.path.join(folder_config["output_folder"], "hub-report.json"),
        json.loads(hub_report),
    )
    print("\t... done!")

    # ---- INFERENCE ----

    print("Loading inference data ...")
    # ingest test data
    inference_data = ConstraintNavigatorData.create()

    files_filter = {"includeFiles": [INFERENCE_DATA_FILE]}
    data_filter = {
        "_type": "Inference",
        "dataSpecifications": [hub_model["inferenceDataSpecification"]],
        "inferenceTimeRanges": [{"startTime": 1749, "endTime": 1797}],
    }
    files_reader_config = json.dumps(
        {
            "folder": folder_config["input_folder"],
            "filesFilter": files_filter,
            "dataFilter": data_filter,
        }
    )

    inference_data.read_files(files_reader_config)

    print("\t... done!")

    print("Creating inference ...")
    # define model constraints
    model_constraints = [
        {
            "model": "sp_model_0",
            "condition": {
                "_type": "StaticNumerical",
                "lowerThreshold": 0.9,
            },
        }
    ]

    # define signal constraints
    # the gray shades of the pixel are a number between 0 and 16
    # therefore we restrict our search space also to this range for all 64 pixels
    signal_constraints = []
    for i in range(0, 64):
        signal_constraints.append(
            {
                "signal": str(i),
                "condition": {
                    "_type": "StaticNumerical",
                    "lowerThreshold": 0.0,
                    "upperThreshold": 16.0,
                },
            }
        )

    # define inference config
    inference_config = json.dumps(
        {
            "skipOnInsufficientData": True,
            "cost": {"_type": "L2Distance"},
            "modelConstraints": model_constraints,
            "signalConstraints": signal_constraints,
            "dev": {
                "penaltyInitial": 25,
            },
        }
    )

    # create inference instance from hub and config
    inference = ConstraintNavigatorInference.create_by_hub(hub, inference_config)
    print("\t... done!")

    print("Calculating next normal values ...")
    # choose inference timestamps: these are the timestmaps for digit 9
    timestamps = list(range(1749, 1797))

    # define next normal config (needed if we want signal filter)
    next_normal_config = json.dumps({})

    # perform constraint optimization
    predictions = inference.infer_float_with_next_normal(
        inference_data, timestamps, next_normal_config
    )
    print("\t... done!")

    print("Save inference file ...")
    predictions_json = float_constraint_values_with_next_normal_to_json(
        predictions, inference_config
    )
    write_json(
        os.path.join(
            folder_config["output_folder"],
            "float-constraint-values-with-next-normal.json",
        ),
        predictions_json,
    )
    print("\t... done!")

    print("Destroy everything ...")
    # destroying can be done once the respective object is not needed anymore - do this if memory is an issue
    model_context.destroy()
    hub.destroy()
    inference_data.destroy()
    inference.destroy()
    print("\t... done!")

    print("Generating report ...")

    # ---- Toolbox HTML Report ----
    from aivis_engine_v2_toolbox.api import build_constraint_navigator_report

    build_constraint_navigator_report(
        {
            "title": "Aivis Engine V2 - Constraint Navigator Example",
            "constraintValuesFile": os.path.join(
                folder_config["output_folder"],
                "float-constraint-values-with-next-normal.json",
            ),
            "hubReportFile": os.path.join(
                folder_config["output_folder"], "hub-report.json"
            ),
            "hubModelFile": os.path.join(
                folder_config["output_folder"], "hub-model.json"
            ),
            "archive": {"folder": folder_config["input_folder"]},
            "outputFile": os.path.join(
                folder_config["output_folder"],
                "aivis-engine-v2-constraint-navigator-example.html",
            ),
        }
    )

    print("ALL DONE!")
    return


# run main function
main()
